/* Copyright  2013 Vermilion Software ltd.  This software is the confidential and proprietary information of Vermilion Software ltd. (Confidential Information). You are not 
 * permitted to use, copy, disclose or broadcast such Confidential Information, save that Vermilion Software Ltd has permitted certain parties to use it on the terms of certain licence 
 * agreements.  If you are not party to such an agreement, you are requested to report the matter to Vermilion Software Ltd on +44 (0)845 241 9966.  Failure to do so may result in a criminal 
 * offence and legal liability to compensate Vermilion Software ltd.  Vermilion Software ltd will rigorously enforce all its rights in relation to the Confidential Information.
 */
package com.vermilion.vrs.dgs.customprocessors;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.TreeMap;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.vermilion.vrs.common.StringHelper;
import com.vermilionsoftware.vrs.common.exception.VermilionFatalException;
import com.vermilionsoftware.vrs.common.exception.VermilionNonFatalException;
import com.vermilionsoftware.vrs.dgs.api.ExternalProcessor;
import com.vermilionsoftware.vrs.dgs.api.ParameterList;
import com.vermilionsoftware.vrs.dgs.api.RecordSet;
import com.vermilionsoftware.vrs.dgs.api.RecordSetList;
import com.vermilionsoftware.vrs.dgs.api.RecordSetMetaData;
import com.vermilionsoftware.vrs.dgs.api.WarningMessage;

/**
 * @author Jill.Heaney
 *
 * This is the abstract DTP that all DTP's should implement.
 * The common set of instructions are implemented in the DragThoroughProcessorInstructions
 * and client specific should be in DragThroughProcessor
 *
 */

public abstract class DragThroughProcessorParent implements ExternalProcessor {
	
	protected static int MAXIMUM_OPERATIONS = 99; //cap on instruction-reading loop
	protected static String DEFAULT_DRAGTHROUGH = "BOTH"; // options are RS_ONLY, DS_ONLY, BOTH
	protected static String INSTRUCTIONS_FIELD = "BOTH"; // options are NOTES_ONLY, INSTRUCTIONS_ONLY, BOTH
	
    protected Log logManager = LogFactory.getLog(this.getClass().getName());
	protected TreeMap<String,String[]> allParams = new TreeMap<String, String[]>(String.CASE_INSENSITIVE_ORDER);
	protected ArrayList<String[]> instructions = new ArrayList<String[]>();
    protected RecordSet outputRecordSet;
    protected RecordSetList inputRecordSetList;
    protected ParameterList parameterList;
	protected String runPos = "Starting";
	protected Properties properties;
	protected boolean trace = false;
	protected boolean invalidInstructionWarnings = true;
	protected boolean allowDefaultDragThrough = true;
	
    protected abstract boolean runInstructions(String key) throws VermilionNonFatalException, VermilionFatalException;


    protected String getParentName() {
        //3.8.7: removed debug statement that processData is starting. fills logs.
        //3.8.8: prevented the auto dragthrough from happening twice
    	//3.8.9: Added ability to include multi-line instructions using the ={ .... } convention; this will concatenate 
    	//       all lines between the ={ and the final } character (with single spaces added between)
    	//       e.g. 
    	//         joinWhere={
    	//           (DS144.BenchmarkID='$_Fund.BenchmarkID$' 
		//            and DS144.ClassGroupNm='$_Classification.Code$' 
		//            and DS142.$_Classification.Code$Nm=DS144.ClassNm) 
		//            OR (DS142.SecurityID='CASH' and DS144.ClassNm='CASH')
		//         }
    	//
    	//         nextInstruction=
    	
        return "p3.8.9"; 
    }
    
    public int preRetrieve(ParameterList parameterList, String notes) {
    	return 0; // 0=retrieve  (1=don't retrieve)
    }

    /**
     * Simple (not anymore!) custom processor that returns the first recordset it finds in the input recordset
     * list - ignoring any datasources that are provided alongside
     *
     * @param  inputRecordSetList List of incoming RecordSets.
     * @param  parameterList List of parameters used in data sources.
     * @param  outputRecordSet The output Record Set.
     * @throws VermilionFatalException
     * @throws VermilionNonFatalException
     */
    public void processData(RecordSetList inputRecordSetList, ParameterList parameterList, RecordSet outputRS) throws VermilionFatalException, VermilionNonFatalException {
    	this.outputRecordSet = outputRS;
    	this.inputRecordSetList = inputRecordSetList;
    	this.parameterList = parameterList;

    	try {
    		copyWarningMessages(inputRecordSetList, outputRecordSet);
    		
    		//Retrieves instructions from the datasource definition fields:
    		if(INSTRUCTIONS_FIELD.equals("BOTH")){
    			setupHelp(outputRecordSet.getDataSourceInfo().getInstructions()+
    					"\n"+outputRecordSet.getDataSourceInfo().getNotes());
    		} else if (INSTRUCTIONS_FIELD.equals("NOTES_ONLY")){
    			setupHelp(outputRecordSet.getDataSourceInfo().getNotes());
    		} else if (INSTRUCTIONS_FIELD.equals("INSTRUCTIONS_ONLY")){
    			setupHelp(outputRecordSet.getDataSourceInfo().getInstructions());
    		} else{
    			setupHelp("");
    		}

            //logManager.debug("processData() starting...");

	        if(instructions.size()==0) { 
	        	writeOutputRecordset(inputRecordSetList);
	        	return;
	        }
	        
            // Loop through the instructions processing them in order
            int currentLoop = 0;
            
    		while(instructions.size() > 0) {

    			if(currentLoop++ > MAXIMUM_OPERATIONS) {
    	    		outputRecordSet.addWarningMessage("Max number of instructions is " + 
    	    				MAXIMUM_OPERATIONS + ". All subsequent instructions ignored");
    				return;
    			}
        		String[] instruction = instructions.get(0);	// always get the top one
    			String key = instruction[0];
        		if(! key.matches("^[a-zA-Z0-9].*$")) {
                    instructions.remove(0);
                    if(trace) 
                    	outputRecordSet.addWarningMessage(new WarningMessage(0, "Rendering",  WarningMessage.DONT_OUTPUT,
                    			"Comment ignored : " + key));
        		} else if ( key.equalsIgnoreCase("QuietMode")){
        			String quietModeString = getAndRemoveInstruction("QuietMode").toUpperCase();
        	        if ( quietModeString.equals("N")) {
        	        	trace = true;
        	        	invalidInstructionWarnings = true;
        	        }else if ( quietModeString.equals("S")) {
        	        	trace = false;
        	        	invalidInstructionWarnings = true;
        	        }else if ( quietModeString.equals("Y")) {
        	        	trace = false;
        	        	invalidInstructionWarnings = false;
        	        } 
        		} else {
        			//Date date= new Date();
        			//Timestamp timestamp = new Timestamp(date.getTime());
        			//logManager.debug("about to run instruction'"+key+"' at "+timestamp);
        			dragThroughDSIfRequired(key);
        			
        			// Instruction Method Developer note:
        			// runInstructions(key) is in the DragThroughProcessor or the DragThroughProcessInstructions;
        			// this method will look at each instruction and decide which process method to call. 
        			// EACH METHOD SHOULD INCLUDE THE RECOLLECTION OF THE INSTRUCTION SO TO REMOVE IT FROM THE LIST
        			// Once the instruction method has run, it will return true to move on, or false to remove the 
        			// instruction, add a warning and then move on
        			
        			if(! runInstructions(key)){
        				//the key did not match any known instruction.
                        instructions.remove(0);
                        if(invalidInstructionWarnings){
                        	outputRecordSet.addWarningMessage(new WarningMessage(0, "Rendering",  WarningMessage.DONT_OUTPUT,
                        		"Unknown instruction: '" + key + "'. Removed. If this is an additional"
                        		+" command for another instruction, please put this after."));
                        }
        			}
        		}
    		}
    	} catch (Exception e) {
    		outputRecordSet.addWarningMessage("Unexpected exception "+e.toString()+" during "+runPos);
    	}
    }

    /**
     * Check if this instruction should be working on the input or the output recordset
     * if it is the output recordset, and that is empty - then copy across the input
     * 
     * @param key
     * @throws VermilionNonFatalException
     */
    private void dragThroughDSIfRequired(String key) throws VermilionNonFatalException {
		String[] value = allParams.get(key);
		if(value==null || value.length==0) {
			return;
		}
		if(value[0].equals("Y") && outputRecordSet.getRowCount()==0 && allowDefaultDragThrough) {
        	writeOutputRecordset(inputRecordSetList);
		}
		allowDefaultDragThrough=false;

    }
    
    /**
     * Create a list of all the possible parameters that could be passed in the notes section,
     * with their default value (default meaning what will be used if that parameter isn't passed),
     * the actual value passed (null if not passed) and the help notes
     * 
     * @throws VermilionNonFatalException
     */
	protected void setupHelp(String dsNotes) throws VermilionNonFatalException {
		runPos = "SetupHelp starting";
		allParams.put("help",		new String[] {"Y","-", "Y|N (will display the help text"});
		allParams.put("quietMode",	new String[] {"S","-", "Y|S|N (Default S. If N it will display messages showing its progress. If S or N, it will display warnings for instructions that are unknown to DTP, such as WF instructions"});

		getHelpText();

		// Load the instructions
		// You can't use a properties file 'cos that will expect the commands to be unique, so we can't have >1 of them
		if(dsNotes==null)
			return;
		
		String lines[] = dsNotes.split("\\r?\\n"); 

		boolean inMultiLine=false;
		String cmd="";
		String[] instruction = new String[2];
		for(String line: lines) {
			if(line!=null) {
				line=line.trim();
				if ( inMultiLine ) {
					if ( line.endsWith("}") ) {
						inMultiLine=false;
						line=line.substring(0, line.length()-1);
					}
					instruction[1]=instruction[1]+" "+line;
					instructions.remove(instructions.size()-1);
					instructions.add(instruction);
				}
				else {
					if ( line.endsWith("={") ) {
						inMultiLine=true;
						line=line.substring(0, line.length()-1);
					}
					
					int i = line.indexOf("=");
					if(i>0) {
						instruction = new String[2];
						cmd = line.substring(0,i);
						instruction[0] = cmd;
						instruction[1] = line.substring(i+1);
						instructions.add(instruction);
					}
				}
			}
		}

		if(getAndRemoveInstruction("help").equalsIgnoreCase("y")) {
	        getHelp(outputRecordSet);
	    }
		runPos = "SetupHelp finished";

	}
	
	protected abstract void getHelpText();
    // -----------------------------------------------------------------------------------------------------------------------
    // These are helper methods, called by primary modules of this class:
    // -----------------------------------------------------------------------------------------------------------------------

	protected String getSplit(String a, String b, int i) {
    	String[] x = a.split(b);
    	return x[i];
    }

	protected void deleteOutputRecordset() throws VermilionNonFatalException {
		runPos = "deleteOutputRecordset starting";
		int rowCount = outputRecordSet.getRowCount();
		
		for (int i=0; i<rowCount; i++) {
			outputRecordSet.removeRow(0);
		}
	}
	
	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void writeOutputRecordset() throws VermilionNonFatalException {
		//Outputs either all input datasources, all input recordsets, or both.
		runPos = "writeOutputRecordset() starting";
		
		boolean outputRS = DEFAULT_DRAGTHROUGH.equals("BOTH") ||  DEFAULT_DRAGTHROUGH.equals("RS_ONLY");
		boolean outputDS = DEFAULT_DRAGTHROUGH.equals("BOTH") ||  DEFAULT_DRAGTHROUGH.equals("DS_ONLY");
		for (int r=0; r<inputRecordSetList.getRecordSetCount(); r++) {
			if ( outputDS && inputRecordSetList.getDataSourceName(r) != null){
				writeOutputRecordset(inputRecordSetList.getRecordSet(r));
			} else if ( outputRS && inputRecordSetList.getRecordSetName(r) != null){
				writeOutputRecordset(inputRecordSetList.getRecordSet(r));
			}
		}
	}
	
	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void writeOutputRecordset(String dsCode) throws VermilionNonFatalException {
		runPos = "writeOutputRecordset(String) starting";

		RecordSet dragThroughRSorDS = this.getInputRecordSet(dsCode, true);
		if(dragThroughRSorDS !=null){
			writeOutputRecordset(dragThroughRSorDS);
		}
		copyWarningMessages(inputRecordSetList, outputRecordSet);
	}
	
	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void writeOutputRecordset(RecordSetList inputRecordSetList) throws VermilionNonFatalException {
		for (int r=0; r<inputRecordSetList.getRecordSetCount(); r++) {
			writeOutputRecordset(inputRecordSetList.getRecordSet(r));
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void writeOutputRecordset(RecordSetList inputRecordSetList, String dsCode) throws VermilionNonFatalException {
		writeOutputRecordset(inputRecordSetList.getRecordSetByDataSourceCode(dsCode));

		copyWarningMessages(inputRecordSetList, outputRecordSet);
	}
	
	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void writeOutputRecordsetUsingColName(String dsCode) throws VermilionNonFatalException {
		runPos = "writeOutputRecordsetUsingColName(String) starting";

		RecordSet dragThroughRSorDS = this.getInputRecordSet(dsCode, true);
		if(dragThroughRSorDS !=null){
			writeOutputRecordsetUsingColName(dragThroughRSorDS);
		}
		copyWarningMessages(inputRecordSetList, outputRecordSet);
	}
	
	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void writeOutputRecordsetUsingColName(RecordSet irs) throws VermilionNonFatalException {
		// we need to map the rows by name
		int rowCount = irs.getRowCount();
		Integer[] outputColumnNumbers = new Integer[irs.getColumnCount()];
		
    	for(int cin=0; cin<irs.getColumnCount(); cin++) {
    		String inputColSysName = irs.getMetaData(cin).getFieldName();
    		String inputColPrefName = irs.getMetaData(cin).getPreferredName();
    		for(int  cout=0; cout<outputRecordSet.getColumnCount(); cout++) {
    			String outputColSysName = outputRecordSet.getMetaData(cout).getFieldName();
    			String outputColPrefName = outputRecordSet.getMetaData(cout).getPreferredName();
    			if(inputColSysName!=null){
    				if(outputColSysName!=null){
    					if(inputColSysName.equals(outputColSysName)){
    						outputColumnNumbers[cin]=cout;
    					}
    				}
    				if(outputColPrefName!=null){
    					if(inputColSysName.equals(outputColPrefName)){
    						outputColumnNumbers[cin]=cout;
    					}   	    					
    				}
    			}
    			if(inputColPrefName!=null){
    				if(outputColSysName!=null){
    					if(inputColPrefName.equals(outputColSysName)){
    						outputColumnNumbers[cin]=cout;
    					}
    				}
    				if(outputColPrefName!=null){
    					if(inputColPrefName.equals(outputColPrefName)){
    						outputColumnNumbers[cin]=cout;
    					}
    				}
    			}
    		}
    	}
    	
     	for(int rec=0; rec<rowCount; rec++) {
     		int newRow = outputRecordSet.addRow();
	    	for(int c=0; c<irs.getColumnCount(); c++) {
	    		if(outputColumnNumbers[c]!=null){
   	    			outputRecordSet.setValue(newRow, outputColumnNumbers[c], irs.getRow(rec).getString(c));
	    		}
	    	}
     	}
     	allowDefaultDragThrough=false;
     	return;
	}
	
	protected void writeOutputRecordset(RecordSet irs) throws VermilionNonFatalException {
		int rowCount = irs.getRowCount();
		runPos="writeOutputRecordset starting";
		RecordSetMetaData rsmdIn = irs.getMetaData();
		RecordSetMetaData rsmdOut = outputRecordSet.getMetaData();
		String inColName;
		String outColName;
		
		if(irs.getColumnCount()==outputRecordSet.getColumnCount()) {
    		// you may have the same number of input and output columns, but if one of the input cols contains the word 'null'
    		// then this turns the column into a null col and it isn't returned, so need to match on column names
			// if your columns names aren't identical then this won't work
			runPos="writeOutputRecordset: using same column count";
	     	for(int inRow=0; inRow<rowCount; inRow++) {
				runPos="writeOutputRecordset: writing row number "+inRow;
	     		if(outputRecordSet.getColumnCount()==irs.getRow(inRow).getColumnCount()) {
		     		outputRecordSet.addRow(irs.getRow(inRow));
	     		} else {
	     			int outRow = outputRecordSet.addRow();
	     			for(int outCol=0; outCol<outputRecordSet.getColumnCount(); outCol++) {
	     				outColName = rsmdOut.getField(outCol).getFieldName();

	     				for(int inCol=0; inCol<irs.getColumnCount(); inCol++) {
		     				inColName = rsmdIn.getField(inCol).getFieldName();

		     				if(inColName.equals(outColName)) {
		     					outputRecordSet.setValue(outRow, outCol,  irs.getString(inRow, inCol));
		     				}
		     			}
	     			}
	     		}
	     	}
	     	return;
		} 
    		
		// if they are different then check if the output RS has more or less cols than the input
		runPos="writeOutputRecordset: using different column count";
		int cols = 0;
		if(irs.getColumnCount() < outputRecordSet.getColumnCount()) {
			cols = irs.getColumnCount();
		} else {
			cols = outputRecordSet.getColumnCount();
		}
		
     	for(int rec=0; rec<rowCount; rec++) {
     		int newRow = outputRecordSet.addRow();
	    	for(int c=0; c<cols; c++) {
	    			outputRecordSet.setValue(newRow, c,  irs.getRow(rec).getString(c));
	    	}
     	}
	}
	
	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void writeCombinedOutputRecordset(RecordSet rst1, RecordSet rst2)  throws VermilionNonFatalException {
		WarningMessage msg = null;
				
		RecordSetMetaData inRsMd1 = rst1.getMetaData();
		RecordSetMetaData inRsMd2 = rst2.getMetaData();
		RecordSetMetaData outRsMd = outputRecordSet.getMetaData();
		
		if ((rst1.getRowCount() > 0) && (rst2.getRowCount() > 0)) {
			//Both input recordsets have data
			for (int inRow1 = 0; inRow1 < rst1.getRowCount(); inRow1++) {
				for (int inRow2 = 0; inRow2 < rst2.getRowCount(); inRow2++) {
					int outRow = outputRecordSet.addRow();
	    	    	
					for (int outCol = 0; outCol < outputRecordSet.getColumnCount(); outCol++) {
						//Check rst1 for a matching column
						int inCol = getColumnIndexBySystemNm(inRsMd1, outRsMd.getField(outCol).getFieldName(), false);
								
						if (inCol != -1) {
							outputRecordSet.setValue(outRow, outCol,  rst1.getRow(inRow1).getString(inCol));
						} else {
							//Check rst2 for a matching column
							inCol = getColumnIndexBySystemNm(inRsMd2, outRsMd.getField(outCol).getFieldName(), false);
							
							if (inCol != -1) {
								outputRecordSet.setValue(outRow, outCol,  rst2.getRow(inRow2).getString(inCol));
							} else {
								msg = new WarningMessage(0, "Rendering",  WarningMessage.DONT_OUTPUT, "Could not find column " + outRsMd.getField(outCol).getFieldName());
								outputRecordSet.addWarningMessage(msg);
							}							
						}						
					}
				}			 
			}
		} else if (rst1.getRowCount() > 0) {
			//Only rst1 has data
			for (int inRow1 = 0; inRow1 < rst1.getRowCount(); inRow1++) {
				int outRow = outputRecordSet.addRow();
	    	    	
				for (int inCol = 0; inCol < rst1.getColumnCount(); inCol++) {
					//Check output recordset for a matching column
					int outCol = getColumnIndexBySystemNm(outRsMd, inRsMd1.getField(inCol).getFieldName(), false);
								
					if (outCol != -1) {
						outputRecordSet.setValue(outRow, outCol,  rst1.getRow(inRow1).getString(inCol));																
					}
				}			 
			}
		} else if (rst2.getRowCount() > 0) {
			//Only rst2 has data
			for (int inRow2 = 0; inRow2 < rst2.getRowCount(); inRow2++) {
				int outRow = outputRecordSet.addRow();
	    	    	
				for (int inCol = 0; inCol < rst2.getColumnCount(); inCol++) {
					//Check output recordset for a matching column
					int outCol = getColumnIndexBySystemNm(outRsMd, inRsMd2.getField(inCol).getFieldName(), false);
								
					if (outCol != -1) {
						outputRecordSet.setValue(outRow, outCol,  rst2.getRow(inRow2).getString(inCol));																
					}
				}			 
			}
		}
	}
	
	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected RecordSet getRecordSet(String code) throws VermilionNonFatalException {
        if ( inputRecordSetList.getDataSourceCodes().contains(code) ) {
        	return inputRecordSetList.getRecordSetByDataSourceCode(code);
        } else {
        	return inputRecordSetList.getRecordSetByRecordSetCode(code);
        }
	}
	
	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
    public String getProperty(String propertiesText, String propertyName, String defaultValue) {
    	// load the properties if not already done
    	if(properties==null) {
            try {
                ByteArrayInputStream reader = new ByteArrayInputStream(propertiesText.getBytes());
                properties = new Properties();
                properties.load(reader);
            } catch (IOException e) {
            	e.printStackTrace();
            }
    	}
    	
        String value = null;
        if (propertiesText == null || propertiesText.length() == 0) {
            return defaultValue;
        } else {
            value = properties.getProperty(propertyName);

            if (value == null) {
                value = defaultValue;
            }
        }
        return value;
    }

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
    //return a record set by name or by code
    protected RecordSet getInputRecordSet(String dsCodeOrName, boolean detailedWarning) throws VermilionNonFatalException{
		try{
			RecordSet requestedRS = inputRecordSetList.getRecordSetByDataSourceCode(dsCodeOrName);
			return requestedRS;
		}catch(VermilionNonFatalException e){
			if(!e.toString().contains("Record set not found"))
				throw e;
		}
		try{
			RecordSet requestedRS = inputRecordSetList.getRecordSetByDataSourceName(dsCodeOrName);
			return requestedRS;
		}catch(VermilionNonFatalException e){
			if(!e.toString().contains("Record set not found"))
				throw e;
		}
		try{
			RecordSet requestedRS = inputRecordSetList.getRecordSetByRecordSetCode(dsCodeOrName);
			return requestedRS;
		}catch(VermilionNonFatalException e){
			if(!e.toString().contains("Record set not found"))
				throw e;
		}
		try{
			RecordSet requestedRS = inputRecordSetList.getRecordSetByRecordSetName(dsCodeOrName);
			return requestedRS;
		}catch(VermilionNonFatalException e){
			if(!e.toString().contains("Record set not found"))
				throw e;
		}
		if(detailedWarning){
			StringBuilder inputCodes = new StringBuilder();
			StringBuilder inputNames = new StringBuilder();

			for (int r=0; r<inputRecordSetList.getRecordSetCount(); r++) {
	      		if ( inputRecordSetList.getDataSourceCode(r) != null) {
	      			inputCodes.append(r + " " + inputRecordSetList.getDataSourceCode(r));
	      		}
	      		if ( inputRecordSetList.getRecordSetCode(r) != null) {
	      			inputCodes.append(r + " " + inputRecordSetList.getRecordSetCode(r));
	      		}
	      		if ( inputRecordSetList.getDataSourceName(r) != null) {
	      			inputNames.append(r + " " + inputRecordSetList.getDataSourceName(r));
	      		}
	      		if ( inputRecordSetList.getRecordSetName(r) != null) {
	      			inputNames.append(r + " " + inputRecordSetList.getRecordSetName(r));
	      		}
	      		
	      		inputCodes.append(" ");
	      		inputNames.append(" ");

			}
			outputRecordSet.addWarningMessage(dsCodeOrName + " not found. The codes available are : " + inputCodes.toString());
			outputRecordSet.addWarningMessage(dsCodeOrName + " not found. The names available are : " + inputNames.toString());

		}
    	return null;
    }

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
    // as we can have the instructions passed multiple times they must be removed as they are processed
    protected String getAndRemoveInstruction(String key) {
    	for(int i=0; i<instructions.size(); i++) {
    		String[] instruction = instructions.get(i);
    		if(instruction==null || instruction.length==0) {
    			outputRecordSet.addWarningMessage("no available instructions");
    			return null;
    		}
    		
			if(instruction[0].equalsIgnoreCase(key)) {
				instructions.remove(i);
				return instruction[1];
			}
		}
    	
        // if it doesn't exist then return the default value for it
		String[] value = allParams.get(key);
		if(value==null) {
			outputRecordSet.addWarningMessage("The help text for instruction " + key + " isn't setup - please correct in the DTP code");
			return "";
		}
		return value[1];
    }

	protected boolean isInstructionValuePassed(String key) {
		boolean passed = false;
		
		for(String[] instruction: instructions) {
			if(instruction[0].equalsIgnoreCase(key)) {
				passed = true;
			}
		}
		return passed;
	}
	
	// as the setRecordSet command will wipe out all of the previous warning messages we need to store them 1st and then add them
	protected void setRecordSet(RecordSet recordSet) throws VermilionNonFatalException {
		int numMessages = outputRecordSet.getWarningMessageCount();
		WarningMessage[] msgs = new WarningMessage[numMessages];

		
		for(int i=0; i<numMessages; i++) {
			msgs[i] = outputRecordSet.getWarningMessage(i);
		}
		
	    outputRecordSet.setRecordSet(recordSet);

		for(int i=0; i<numMessages; i++) {
			outputRecordSet.addWarningMessage(msgs[i]);
		}
	}

	   
    /**
     * getHelp used to provide help information on request so that GUI users can
     * see the processor Notes options
     */
    public void getHelp(RecordSet outputRecordSet, ArrayList<String> help) {
        WarningMessage warningMessage = new WarningMessage(0, "Help", WarningMessage.DONT_OUTPUT, getRevision());
        outputRecordSet.addWarningMessage(warningMessage);
        for (int h = 0; h < help.size(); h++) {
            warningMessage = new WarningMessage(0, "Help", WarningMessage.DONT_OUTPUT, (String) help.get(h));
            outputRecordSet.addWarningMessage(warningMessage);
        }
    }
		
    public void getHelp(RecordSet outputRecordSet) {
        WarningMessage warningMessage = new WarningMessage(0, "Help", WarningMessage.DONT_OUTPUT, getRevision());
        outputRecordSet.addWarningMessage(warningMessage);
        
        Map<String, String[]> treeMap = new TreeMap<String, String[]>(allParams);

        Collection<String> c = treeMap.keySet();

        Iterator<String> itr = c.iterator();
        while(itr.hasNext()) {
        	String key = itr.next();
        	String[] value = allParams.get(key);
        	String helpMsg = key + "=" + value[2] + "; default = " + value[1] + ")";
            warningMessage = new WarningMessage(0, "Help", WarningMessage.DONT_OUTPUT, helpMsg);
            outputRecordSet.addWarningMessage(warningMessage);
        }
    }
    
    protected String getSVWarning(String fieldVal, ParameterList parameterList) throws VermilionFatalException, VermilionNonFatalException {
        boolean noted=false;
        if ( fieldVal.indexOf("$_") > -1 ) {
            for (int p=0; p<parameterList.getParameterCount(); p++) {
                if ( fieldVal.indexOf("$_"+parameterList.getParameterRecord(p).getParameterName()+".") > -1 ) {
                    return "Did not resolve '"+ StringHelper.getResolveSubVars( fieldVal, parameterList, true)
							    	+"'; check fields (available = "
							    	+parameterList.getParameterRecord(p).getFieldNames()+"="+fieldVal.indexOf("$_"+parameterList.getParameterRecord(p).getParameterName()+".")+") (err10002)";
                }
                else {
                    noted=true;
                }
			}
            if ( ! noted ) {
                return "Did not resolve '"+ StringHelper.getResolveSubVars( fieldVal, parameterList, true)+"'; check entities (available = "+parameterList.getParameterNames()+"; (err10001)";
            }
        }
        return "";      
    }

    // ------------------------------------------------------------------------------------------------------------------------------
	/**
     * Find the column name passed into the method in the recordset that this passed 
     * in and return the positional index to the caller. If the column name is not found
     * the method will return -1.
     * 
	 * @param _inputRSMD metadata for the RecordSet
	 * @param colName name of the column we want to know the index of
	 * @param showWarnings switch to display warnings when columns can't be found
	 * 
	 * @return Position of the column name in the recordset
	 */
	protected int getColumnIndexBySystemNm(RecordSetMetaData _inputRSMD, String _colName, boolean showWarnings) {
	    if ( _inputRSMD == null ) {
	        logManager.error("getColumnIndexBySystemNm called with RecordSetMetaData = null - please investigate");
	        return -1;
	    }
	    int iCols = -1;
	    try {
	         iCols = _inputRSMD.getColumnCount();
	    }
	    catch (Exception e) {
	        logManager.error("getColumnIndexBySystemNm called with RecordSetMetaData.getColumnCount() and got "+e.toString()+" please investigate (poss recordset has no records)");
	        return -1;
	    }
		int iPos = -1;
		String sCurrName = null;
		StringBuilder availableCols = new StringBuilder();
		
		for (int i=0; i<iCols; i++) {
			sCurrName = _inputRSMD.getField(i).getFieldName();
			availableCols.append(sCurrName + ", ");
			// logManager.error("Looking at "+i+" : "+sCurrName+" (vs. "+_colName+")");
			if ( (sCurrName != null) && _colName.equalsIgnoreCase(sCurrName) ) {
				iPos = i;
				return iPos;
			} 
		}
		if ( iPos == -1 && showWarnings) {
			String msg = "Unable to find the resultCol : " + _colName + ". Available cols are : " + availableCols.toString();
			outputRecordSet.addWarningMessage(msg);
		}
		
		return iPos;
	}

    // ------------------------------------------------------------------------------------------------------------------------------
	protected int getColumnIndexBySystemNm(RecordSetMetaData _inputRSMD, String _colName) {
    	return getColumnIndexBySystemNm(_inputRSMD, _colName, true);
    }

    // ------------------------------------------------------------------------------------------------------------------------------
	protected void copyWarningMessages(RecordSetList inputRecordSetList, RecordSet outputRecordSet) throws VermilionNonFatalException {
		    
 	    ArrayList<String> warnings = new ArrayList<String>();
 	    ArrayList<String> cautions = new ArrayList<String>();

 	    int warningCount = outputRecordSet.getWarningMessageCount();
 		for(int w=0; w<warningCount; w++) {
 	        warnings.add(outputRecordSet.getWarningMessageText(w));
 	    }
 		
 		for(int l=0; l<inputRecordSetList.getRecordSetCount(); l++) {
 			RecordSet inputRecordSet = inputRecordSetList.getRecordSet(l);
 			warningCount = inputRecordSet.getWarningMessageCount();
 			int renderingCount = inputRecordSet.getRenderingInstructionCount();
 			int cautionCount = inputRecordSet.getCautionMessageCount();

 			if (cautionCount > 0) {
 				for (int c = 0 ; c < cautionCount ; c++) {
 					if ( ! cautions.contains( inputRecordSet.getCautionMessageText(c) ) ) {
 					    outputRecordSet.addCautionMessage(inputRecordSet.getCautionMessage(c));
 					    cautions.add(inputRecordSet.getCautionMessageText(c));
 					}
 				}
 	        }                

            // Copy any warning messages from the input record set into the output record set.
 			if (warningCount > 0) {
 				for (int warningIndex = 0 ; warningIndex < warningCount ; warningIndex++) {
 					//customLog("[copyWarningMessages] message text = " + inputRecordSet.getWarningMessage(warningIndex).getMessageText());
 					if ( ! warnings.contains( inputRecordSet.getWarningMessageText(warningIndex) ) ) {
 					    outputRecordSet.addWarningMessage(inputRecordSet.getWarningMessage(warningIndex));
 					    warnings.add(inputRecordSet.getWarningMessageText(warningIndex));
 					}
 				}
 	        }                
 	    
 	        if (renderingCount > 0) {
 	            for (int renderingIndex = 0 ; renderingIndex < renderingCount ; renderingIndex++) {
 					//customLog("[copyWarningMessages] rendering instruction xml = " + inputRecordSet.getRenderingInstruction(renderingIndex).getInstructionXML());
 	            	outputRecordSet.addRenderingInstruction(inputRecordSet.getRenderingInstruction(renderingIndex));
 	            }
 	        }
 	     }

 	}
	

	   // ------------------------------------------------------------------------------------------------------------------------------
	   public void copyWarningMessages(RecordSet inputRecordSet, RecordSet outputRecordSet) throws VermilionNonFatalException {
		    
		    ArrayList<String> warnings = new ArrayList<String>();
		    int warningCount = outputRecordSet.getWarningMessageCount();
			for(int w=0; w<warningCount; w++) {
		        warnings.add(outputRecordSet.getWarningMessageText(w));
		    }
			
			warningCount = inputRecordSet.getWarningMessageCount();
			int renderingCount = inputRecordSet.getRenderingInstructionCount();

	        // Copy any warning messages from the input record set into the output record set.
			if (warningCount > 0) {
				for (int warningIndex = 0 ; warningIndex < warningCount ; warningIndex++) {
					//customLog("[copyWarningMessages] message text = " + inputRecordSet.getWarningMessage(warningIndex).getMessageText());
					if ( ! warnings.contains( inputRecordSet.getWarningMessageText(warningIndex) ) ) {
						outputRecordSet.addWarningMessage(inputRecordSet.getWarningMessage(warningIndex));
						warnings.add(inputRecordSet.getWarningMessageText(warningIndex));
					}
				}
		    }                
		    
		    if (renderingCount > 0) {
		    	for (int renderingIndex = 0 ; renderingIndex < renderingCount ; renderingIndex++) {
					//customLog("[copyWarningMessages] rendering instruction xml = " + inputRecordSet.getRenderingInstruction(renderingIndex).getInstructionXML());
		            outputRecordSet.addRenderingInstruction(inputRecordSet.getRenderingInstruction(renderingIndex));
		        }
		    }
		    
		}
	//Automatically updated by SVN
		
	public String getRevision() {
    	return "$Author$, $Date$"; 
    }
	
}
